package edu.sfedu.brsback.services.implementations;

import edu.sfedu.brsback.dto.module.ModuleFilter;
import edu.sfedu.brsback.entities.*;
import edu.sfedu.brsback.entities.Module;
import edu.sfedu.brsback.exceptions.AccessDeniedException;
import edu.sfedu.brsback.exceptions.ModelNotFoundException;
import edu.sfedu.brsback.repos.ModuleRepository;
import edu.sfedu.brsback.repos.RecordBookRepository;
import edu.sfedu.brsback.repos.ScoreRepository;
import edu.sfedu.brsback.services.ModuleRightService;
import edu.sfedu.brsback.services.ModuleService;
import edu.sfedu.brsback.services.SecurityService;
import lombok.RequiredArgsConstructor;
import org.jetbrains.annotations.Contract;
import org.jetbrains.annotations.NotNull;
import org.springframework.stereotype.Service;

import java.util.*;
import java.util.function.Predicate;

@Service
@RequiredArgsConstructor
public class ModuleServiceImpl implements ModuleService {
    private final SecurityService securityService;

    private final ModuleRepository moduleRepository;
    private final ScoreRepository scoreRepository;
    private final RecordBookRepository recordBookRepository;
    private final ModuleRightService moduleRightService;

    @Override
    public List<Module> getAvaliableModules(ModuleFilter filter) {
        return moduleRepository.findAll()
                .stream()
                .filter(genModuleFilterPredicate(filter))
                .toList();
    }

    @Contract(pure = true)
    private @NotNull Predicate<Module> genModuleFilterPredicate(ModuleFilter filter){
        return (Module module) -> {
            var userRight = moduleRightService.getRightOnModule(
                    module,
                    securityService.getAuthUser()
            );
            boolean result = userRight.isRRead();

            if (Boolean.TRUE.equals(filter.getOnlyParents())){
                result &= module.getParent() == null;
            }

            if (filter.getParentModuleId() != null){
                result &= module.getParent() != null && Objects.equals(module.getParent().getId(), filter.getParentModuleId());
            }

            if (filter.getRecordBookId() != null){
                var recordBookRight = moduleRightService.getRightOnModule(
                        module,
                        recordBookRepository.findById(filter.getRecordBookId()).orElseThrow(ModelNotFoundException::new)
                );
                result &= recordBookRight.isRRead();
            }
            if (filter.getTitle() != null){
                result &= module.getTitle().contains(filter.getTitle());
            }
            if (filter.getEducationPeriodId() != null){
                result &= Objects.equals(module.getEducationPeriod().getId(), filter.getEducationPeriodId());
            }

            return result;
        };
    }

    @Override
    public Module getModuleById(Long moduleId) {
        Module module = moduleRepository.findById(moduleId)
                .orElseThrow(ModelNotFoundException::new);

        var userRight = moduleRightService.getRightOnModule(
                module,
                securityService.getAuthUser()
        );

        if (!userRight.isRRead()){
            throw new AccessDeniedException();
        }

        return module;
    }

    @Override
    public void saveModule(Module module) {
        var userRight = moduleRightService.getRightOnModule(
                moduleRepository.findById(module.getId()).orElseThrow(ModelNotFoundException::new),
                securityService.getAuthUser()
        );
        if (
                (module.getId() != null && userRight.isREdit()) ||
                (module.getId() == null && userRight.isRCreate())
        ){
            throw new AccessDeniedException();
        }

        moduleRepository.save(module);
    }

    @Override
    public boolean deleteModuleById(Long moduleId) {
        var userRight = moduleRightService.getRightOnModule(
                moduleRepository.findById(moduleId).orElseThrow(ModelNotFoundException::new),
                securityService.getAuthUser()
        );

        if (!userRight.isRDelete()){
            throw new AccessDeniedException();
        }
        return false;
    }
}
